import unittest
from datetime import datetime
from os import path

import requests
from lxml import etree

from fptest.tracing import parse_trace_file


class FpTest(unittest.TestCase):
    @classmethod
    def zero_file(cls, filename):
        """
        Zero out a file.  Used to clear cartOrderTracing and kpsaOrderTracing before running the tests
        :param filename: File to clear
        """
        with open(filename, 'w'):
            pass

    @classmethod
    def linearize_xml(cls, xml_string):
        """
        Strips insignificant whitespace and adds the xml declaration (unpretty-print)

        This is because FP needs the input to be in this format
        :param xml_string: Pretty-printed xml string
        :return: Headered, unpretty-printed xml string
        """
        parser = etree.XMLParser(remove_blank_text=True)
        elem = etree.XML(xml_string, parser=parser)
        return etree.tostring(elem, xml_declaration=True)

    @classmethod
    def readfile(cls, filename):
        with open(filename, 'r') as f:
            return f.read()

    @classmethod
    def now(cls):
        """
        Get the current timestamp in the format yyyyMMddHHmmssSSSSSS to be used to generate unique orderIds
        :return: timestamp
        """
        return datetime.now().strftime("%Y%m%d%H%M%S%f")

    def __init__(self, *args, **kwargs):
        super(FpTest, self).__init__(*args, **kwargs)
        self.fp_url = 'http://localhost:55000/aff'
        self.fp_node_dir = '../runtime/FPNode'
        self.response = None
        self.cart_order_tracing = None
        self.kpsa_order_tracing = None
        self.fp_response = None
        self.has_run = False

    def request(self):
        """
        Gets the request to post to FP for the test.  There is the helper class method readfile
        :return: XML String of the request
        """
        raise NotImplementedError('The request() method must return the request to send to FP')

    def __do_work(self):
        self.has_run = True
        self.zero_file(path.join(self.fp_node_dir, 'cartOrderTracing.00000.log'))
        self.zero_file(path.join(self.fp_node_dir, 'kpsaOrderTracing.00000.log'))

        data = self.linearize_xml(self.request())
        self.response = requests.post(self.fp_url, data=data)
        self.response.connection.close()
        self.fp_response = etree.XML(self.response.content)
        self.cart_order_tracing = parse_trace_file(path.join(self.fp_node_dir, 'cartOrderTracing.00000.log'))
        self.kpsa_order_tracing = parse_trace_file(path.join(self.fp_node_dir, 'kpsaOrderTracing.00000.log'))

    def setUp(self):
        if not self.has_run:
            self.__do_work()

    def get_fp_status(self):
        """
        Get the execStatus of the response XML
        :return: execStatus
        """
        return self.fp_response.find('execStatus').text

    def get_first_wo(self, wo_name=None):
        """
        Get the first workorder from the list of outgoing work orders in cart_order_tracing
        :param wo_name: Name to query
        :return: Work Order
        """
        if wo_name is None:
            return self.cart_order_tracing.outgoing_workorders[0]
        else:
            return next(wo for wo in self.cart_order_tracing.outgoing_workorders if wo.name == wo_name)



import re
import enum

__author__ = 'John Oxley'


class Trace:
    def __init__(self):
        self.outgoing_workorders = []
        self.lines = []


class WorkOrder:
    @staticmethod
    def balanced_braces(args):
        parts = []
        for arg in args:
            if '<' not in arg:
                continue
            chars = []
            n = 0
            for c in arg:
                if c == '<':
                    if n > 0:
                        chars.append(c)
                    n += 1
                elif c == '>':
                    n -= 1
                    if n > 0:
                        chars.append(c)
                    elif n == 0:
                        parts.append(''.join(chars).lstrip().rstrip())
                        chars = []
                elif n > 0:
                    chars.append(c)
        return parts

    def __init__(self, name=None, status=None):
        self.name = name
        self.status = status
        self.params = {}
        self.raw_params = {}

    def __str__(self):
        return "WO %s: %s" % (self.name, self.status)

    def add_param(self, name, values):
        if values is None:
            self.params[name] = []
        else:
            raw_values = WorkOrder.balanced_braces([values])
            self.raw_params[name] = raw_values
            decoded_values = [v.encode('latin1').decode('unicode_escape') for v in raw_values]
            self.params[name] = decoded_values


class ParserState(enum.Enum):
    looking_for_start = 1
    parsing_outgoing_wo = 2


def parse_trace_file(filename):
    """
    Parse a trace file, e.g. cartOrderTracing.00000.log
    :param filename:
    :return:
    """

    trace = Trace()

    outgoing_workorder = re.compile(r"^(?P<time>.{26}).+OUTGOING: For WO <(?P<wo>[^>]+)>.+<(?P<status>[^>]+)")
    param_re = re.compile(r"^Found param:<(?P<name>[^>]+)>, with value list : ?(?P<values>.+)?")

    parser_state = ParserState.looking_for_start

    wo = WorkOrder()
    with open(filename, 'r') as f:
        for line in f:
            trace.lines.append(line)
            if parser_state == ParserState.looking_for_start:
                wo_match = outgoing_workorder.match(line)
                if wo_match is not None:
                    wo.name = wo_match.group('wo')
                    wo.status = wo_match.group('status')
                    parser_state = ParserState.parsing_outgoing_wo
            elif parser_state == ParserState.parsing_outgoing_wo:
                match = param_re.match(line)
                if match is None:
                    trace.outgoing_workorders.append(wo)
                    wo = WorkOrder()
                    parser_state = ParserState.looking_for_start
                else:
                    wo.add_param(match.group('name'), match.group('values'))

        if parser_state == ParserState.parsing_outgoing_wo:
            trace.outgoing_workorders.append(wo)

    return trace


from os import path

from setuptools import setup


here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='fptest',
    version='0.1.1',
    description='A TIBCO Fulfilment Provisioning test helper',
    long_description=long_description,
    packages=['fptest'],
    url='https://github.com/oxo42/FpTest',
    author='John Oxley',
    author_email='john.oxley@gmail.com',
    license='Apache',

    classifiers=[
        # How mature is this project? Common values are
        # 3 - Alpha
        # 4 - Beta
        # 5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Testing',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        # 'Programming Language :: Python :: 2',
        # 'Programming Language :: Python :: 2.6',
        # 'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],

    requires=['lxml', 'requests']
)
